<?php
/**
 * Perforce Swarm
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Api\Controller;

use Api\AbstractApiController;
use Zend\View\Model\JsonModel;

/**
 * Swarm Servers API
 *
 * @SWG\Resource(
 *   apiVersion="v9",
 *   basePath="/api/v9/"
 * )
 */
class ServersController extends AbstractApiController
{
    /**
     * @SWG\Api(
     *     path="servers/",
     *     @SWG\Operation(
     *         method="GET",
     *         summary="gets a list of servers",
     *         notes="gets a list of servers",
     *         nickname="Servers"
     *     )
     * )
     *
     * @apiUsageExample Servers list endpoint
     *
     *   Get a list of servers that Swarm is aware of
     *
     *   ```bash
     *   curl -u "username:password" \
     *        "https://myswarm.url/api/v9/servers/"
     *   ```
     *
     *   If multiple p4d are configured it will give an output like
     *
     *   ```json
     *   {
     *     "servers": {
     *       "Master": {
     *          "port": "ssl:10.33.44.55:1666"
     *       },
     *       "Artifacts": {
     *          "port": "10.55.66.77:1666"
     *       }
     *     }
     *   }
     *   ```
     *
     *   If a single p4d is configured it will give an output like
     *
     *   ```json
     *   {
     *     "servers": {
     *       "p4": {
     *          "port": "ssl:10.33.44.55:1666"
     *       }
     *     }
     *   }
     *   ```
     *
     * @apiSuccessExample Successful Response:
     *     HTTP/1.1 200 OK
     *
     *     {
     *       "servers": {
     *         "Master": {
     *          "port": "ssl:10.33.44.55:1666"
     *         },
     *         "Artifacts": {
     *          "port": "10.55.66.77:1666"
     *         }
     *       }
     *     }
     *
     * @return JsonModel
     */
    public function getList()
    {
        $services = $this->services;
        $config   = $services->get('config');
        $servers  = array();
        foreach (array_filter(
            isset($config['p4']['port']) ? array($config['p4']) : $config['p4'],
            function ($p4d) {
                return isset($p4d['port']);
            }
        ) as $id => $server) {
            unset($server['user']);
            unset($server['password']);
            // If this is a single server environment return consistent results to
            // multi-server by using 'p4' as the identifier
            $servers[$id === 0 ? 'p4' : $id] = $server;
        }
        return $this->prepareSuccessModel(new JsonModel(array('servers'=> $servers)));
    }
}
