<?php
/**
 * Manager to help with getting values from a configuration array.
 *
 * @copyright   2013-2019 Perforce Software. All rights reserved.
 * @license     Please see LICENSE.txt in top-level readme folder of this distribution.
 * @version     2019.1/1829459
 */

namespace Application\Config;

use Users\Settings\ReviewPreferences;
use Users\Settings\TimePreferences;
use Reviews\Model\Review;
use Workflow\Model\IWorkflow;
use Zend\ModuleManager\Listener\ConfigMergerInterface;
use Markdown\Settings as MarkdownSettings;

class ConfigManager
{
    const ENVIRONMENT                             = 'environment';
    const LOGOUT_URL                              = 'logout_url';
    const VENDOR                                  = 'vendor';
    const EMOJI_PATH                              = 'emoji_path';
    const ENVIRONMENT_LOGOUT_URL                  = 'environment.logout_url';
    const ENVIRONMENT_EMOJI_URL                   = 'environment.vendor.emoji_path';
    const P4                                      = 'p4';
    const SSO_ENABLED                             = 'sso_enabled';
    const P4_SSO_ENABLED                          = 'p4.sso_enabled';
    const DIFF_MAX_DIFFS                          = 'diffs.max_diffs';
    const FILES_DOWNLOAD_TIMEOUT                  = 'files.download_timeout';
    const FILES_MAX_SIZE                          = 'files.max_size';
    const REVIEWS_EXPAND_ALL                      = 'reviews.expand_all_file_limit';
    const REVIEWS_DISABLE_TESTS_ON_APPROVE_COMMIT = 'reviews.disable_tests_on_approve_commit';
    const REVIEWS_DISABLE_APPROVE_WHEN_TASKS_OPEN = 'reviews.disable_approve_when_tasks_open';
    const REVIEWS_EXPAND_GROUP_REVIEWERS          = 'reviews.expand_group_reviewers';
    const REVIEWS_FILTERS_RESULT_SORTING          = 'reviews.filters.result_sorting';
    const REVIEWS_FILTERS_DATE_FIELD              = 'reviews.filters.date_field';
    const UPGRADE_BATCH_SIZE                      = 'upgrade.batch_size';
    const UPGRADE_STATUS_REFRESH_INTERVAL         = 'upgrade.status_refresh_interval';
    const AVATARS_HTTP                            = 'avatars.http_url';
    const AVATARS_HTTPS                           = 'avatars.https_url';
    const SECURITY_REQUIRE_LOGIN                  = 'security.require_login';
    const WORKER_CHANGE_SAVE_DELAY                = 'queue.worker_change_save_delay';
    const TRANSLATOR_NON_UTF8_ENCODINGS           = 'translator.non_utf8_encodings';
    const DASHBOARD_MAX_ACTIONS                   = 'users.maximum_dashboard_actions';
    const DASHBOARD_REFRESH_INTERVAL              = 'users.dashboard_refresh_interval';
    const DISPLAY_FULLNAME                        = 'users.display_fullname';
    const PROJECTS_README_MODE                    = 'projects.readme_mode';
    const PROJECTS_MAX_README_SIZE                = 'projects.max_readme_size';
    const PROJECTS_FETCH_MAXIMUM                  = 'projects.fetch.maximum';
    const PROJECTS_MAINLINES                      = 'projects.mainlines';
    const COMMENT_SHOW_ID                         = 'comments.show_id';
    const COMMENT_THREADING_MAX_DEPTH             = 'comments.threading.max_depth';
    const COMMENT_NOTIFICATION_DELAY_TIME         = 'comments.notification_delay_time';
    const JIRA_LINK_TO_JOBS                       = 'jira.link_to_jobs';
    const JIRA_DELAY_JOB_LINKS                    = 'jira.delay_job_links';
    const JIRA_MAX_JOB_FIXES                      = 'jira.max_job_fixes';
    const JIRA_HOST                               = 'jira.host';
    const JIRA_USER                               = 'jira.user';
    const JIRA_PASSWORD                           = 'jira.password';
    const JIRA_JOB_FIELD                          = 'jira.job_field';
    const REVIEWS_PROCESS_SHELF_DELETE_WHEN       = 'reviews.process_shelf_delete_when';
    const MORE_CONTEXT_LINES                      = 'more_context_lines';
    const REVIEWS_MORE_CONTEXT_LINES              = 'reviews.more_context_lines'; // 5.3 cannot concat const
    const MAX_BOTTOM_CONTEXT_LINES                = 'max_bottom_context_lines';
    const REVIEWS_MAX_BOTTOM_CONTEXT_LINES        = 'reviews.max_bottom_context_lines'; // 5.3 cannot concat const
    const REVIEWS_ALLOW_AUTHOR_OBLITERATE         = 'reviews.allow_author_obliterate';
    const REVIEWS_DISABLE_SELF_APPROVE            = 'reviews.disable_self_approve';
    const REVIEWS_DISABLE_COMMIT                  = 'reviews.disable_commit';
    const MODERATOR_APPROVAL                      = 'moderator_approval';
    const REVIEWS_MODERATOR_APPROVAL              = 'reviews.moderator_approval'; // 5.3 cannot concat const
    const END_STATES                              = 'end_states'; // 5.3 cannot concat const
    const REVIEWS_END_STATES                      = 'reviews.end_states';
    const SECURITY_PREVENT_LOGIN                  = 'security.prevent_login';
    const SECURITY_HTTPS_STRICT                   = 'security.https_strict';
    const MENTIONS_USERS_BLACKLIST                = 'mentions.usersBlacklist';
    const LOG                                     = 'log';
    const REFERENCE_ID                            = 'reference_id';
    const LOG_REFERENCE_ID                        = self::LOG . '.' . self::REFERENCE_ID;
    const EVENT_TRACE                             = 'event_trace';
    const LOG_EVENT_TRACE                         = self::LOG . '.' . self::EVENT_TRACE;
    const MARKDOWN                                = 'markdown';
    const FILE_EXTENSIONS                         = 'file_extensions';
    const MARKDOWN_FILE_EXTENSIONS                = self::MARKDOWN . '.' . self::FILE_EXTENSIONS;
    const MARKDOWN_MARKDOWN                       = self::MARKDOWN . '.' . self::MARKDOWN;

    // User setting.
    const USER_SETTINGS = 'users.settings';
    // User preferences settings.
    const USER_SETTINGS_REVIEW_PREF_SHOW_COMMENTS     = "users.settings.review_preferences.show_comments_in_files";
    const USER_SETTINGS_REVIEW_PREF_SIDE_BY_SIDE      = "users.settings.review_preferences.view_diffs_side_by_side";
    const USER_SETTINGS_REVIEW_PREF_IGNORE_WHITESPACE = "users.settings.review_preferences.ignore_whitespace";
    const USER_SETTINGS_TIME_DISPLAY                  = 'users.settings.time.display';
    // Due to php5.3 not liking constants being put together in a new constant.
    const USER_SETTINGS_REVIEW_PREF_SHOW_SPACE = "users.settings.review_preferences.show_space_and_new_line_characters";

    const SETTINGS = 'settings';

    // Workflow
    // Convenient path prefixes
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW    = 'workflow_rules.on_submit.with_review';
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW = 'workflow_rules.on_submit.without_review';
    const WORKFLOW_END_RULES_UPDATE         = 'workflow_rules.end_rules.update';
    const WORKFLOW_AUTO_APPROVE             = 'workflow_rules.auto_approve';
    // Full workflow paths
    const WORKFLOW_ENABLED                       = 'workflow.enabled';
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW_RULE    = 'workflow_rules.on_submit.with_review.rule';
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_RULE = 'workflow_rules.on_submit.without_review.rule';
    const WORKFLOW_ON_SUBMIT_WITH_REVIEW_MODE    = 'workflow_rules.on_submit.with_review.mode';
    const WORKFLOW_ON_SUBMIT_WITHOUT_REVIEW_MODE = 'workflow_rules.on_submit.without_review.mode';
    const WORKFLOW_END_RULES_UPDATE_RULE         = 'workflow_rules.end_rules.update.rule';
    const WORKFLOW_END_RULES_UPDATE_MODE         = 'workflow_rules.end_rules.update.mode';
    const WORKFLOW_AUTO_APPROVE_RULE             = 'workflow_rules.auto_approve.rule';
    const WORKFLOW_AUTO_APPROVE_MODE             = 'workflow_rules.auto_approve.mode';
    const WORKFLOW_COUNTED_VOTES_RULE            = 'workflow_rules.counted_votes.rule';
    const WORKFLOW_COUNTED_VOTES_MODE            = 'workflow_rules.counted_votes.mode';

    const SAML_HEADER = 'saml.header';

    const HTTP_STRING      = 'httpString';
    const STRING           = 'string';
    const INT              = 'int';
    const BOOLEAN          = 'boolean';
    const ARRAY_OF_STRINGS = 'arrayOfStrings';
    const SWARM_SETTING    = 'swarmSetting';
    const TYPE             = 'type';
    const VALID_VALUES     = 'valid_values';
    const ALLOW_NULL       = 'allowNull';
    // For an array of strings we can define metadata that forces string conversion so that an array such
    // as array('1', 2) will be allowed
    const FORCE_STRING = 'force_string';

    // Case sensitivity levels
    const CASE_SENSITIVITY = 'case_sensitivity';
    const CASE_INSENSITIVE = 1;
    const CASE_SENSITIVE   = 0;

    // Some valid value constants
    const VALUE_ANY  = 'any';
    const VALUE_EACH = 'each';

    // Permitted valid values for the SWARM_SETTING type
    private static $swarmSettings = array(
        Setting::FORCED_DISABLED,
        Setting::FORCED_ENABLED,
        Setting::DISABLED,
        Setting::ENABLED
    );

    // Metadata to define types etc (structure should agree with the configuration array)
    private static $configMetaData = array(
        'comments' => array(
            'threading' => array(
                'max_depth' => array(self::TYPE => self::INT)
            ),
            'show_id'                 => array(self::TYPE => self::BOOLEAN),
            'notification_delay_time' => array(self::TYPE => self::INT)
        ),
        'diffs'   => array(
            'max_diffs'             => array(self::TYPE => self::INT),
        ),
        'files'   => array(
            'max_size'              => array(self::TYPE => self::INT),
            'download_timeout'      => array(self::TYPE => self::INT)
        ),
        'reviews' => array(
            'expand_all_file_limit' => array(self::TYPE => self::INT),
            'filters'               => array(
                'result_sorting'    => array(self::TYPE => self::BOOLEAN),
                'date_field'        =>
                    array(self::TYPE => self::STRING, self::VALID_VALUES => array('created', 'updated'))
            ),
            'expand_group_reviewers'          => array(self::TYPE => self::BOOLEAN),
            'disable_tests_on_approve_commit' => array(self::TYPE => self::BOOLEAN),
            'disable_approve_when_tasks_open' => array(self::TYPE => self::BOOLEAN),
            'process_shelf_delete_when'       => array(
                self::TYPE                 => self::ARRAY_OF_STRINGS,
                self::VALID_VALUES         => array(
                    Review::STATE_APPROVED,
                    Review::STATE_APPROVED_COMMIT,
                    Review::STATE_ARCHIVED,
                    Review::STATE_NEEDS_REVIEW,
                    Review::STATE_NEEDS_REVISION,
                    Review::STATE_REJECTED
                ),
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ),
            self::MORE_CONTEXT_LINES       => array(self::TYPE => self::INT),
            self::MAX_BOTTOM_CONTEXT_LINES => array(self::TYPE => self::INT),
            'allow_author_obliterate'      => array(self::TYPE => self::BOOLEAN),
            'disable_self_approve'         => array(self::TYPE => self::BOOLEAN),
            'disable_commit'               => array(self::TYPE => self::BOOLEAN),
            self::MODERATOR_APPROVAL       =>
                array(self::TYPE  => self::STRING, self::VALID_VALUES => array(self::VALUE_ANY, self::VALUE_EACH)),
            self::END_STATES      => array(
                self::TYPE                 => self::ARRAY_OF_STRINGS,
                self::VALID_VALUES         => array(
                    Review::STATE_APPROVED,
                    Review::STATE_APPROVED_COMMIT,
                    Review::STATE_ARCHIVED,
                    Review::STATE_REJECTED
                ),
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ),
        ),
        'projects' => array(
            'readme_mode' =>
                array(
                    self::TYPE  => self::STRING, self::VALID_VALUES => MarkdownSettings::LEGACY_MARKDOWN_OPTIONS
                ),
            'mainlines'       => array(self::TYPE => self::ARRAY_OF_STRINGS),
            'max_readme_size' => array(self::TYPE => self::INT, self::ALLOW_NULL => true),
            'fetch' => array(
                'maximum' => array(self::TYPE => self::INT)
            )
        ),
        'upgrade' => array(
            'status_refresh_interval' => array(self::TYPE => self::INT),
            'batch_size'              => array(self::TYPE => self::INT)
        ),
        'avatars' => array(
            'http_url'  => array(
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            ),
            'https_url' => array(
                self::TYPE => self::STRING,
                self::CASE_SENSITIVITY => self::CASE_SENSITIVE
            )
        ),
        'security' => array(
            'require_login' => array(self::TYPE => self::BOOLEAN),
            'prevent_login' => array(self::TYPE => self::ARRAY_OF_STRINGS, self::FORCE_STRING => true),
            'https_strict'  => array(self::TYPE => self::BOOLEAN)
        ),
        'queue' => array(
            'worker_change_save_delay' => array(self::TYPE => self::INT)
        ),
        'translator' => array(
            'non_utf8_encodings' => array(self::TYPE => self::ARRAY_OF_STRINGS)
        ),
        'users' => array(
            'maximum_dashboard_actions' => array(self::TYPE => self::INT),
            self::SETTINGS => array(
                ReviewPreferences::REVIEW_PREFERENCES => array(
                    ReviewPreferences::SHOW_COMMENTS_IN_FILES  => array(self::TYPE => self::BOOLEAN),
                    ReviewPreferences::VIEW_DIFFS_SIDE_BY_SIDE => array(self::TYPE => self::BOOLEAN),
                    ReviewPreferences::SHOW_SPACE_AND_NEW_LINE => array(self::TYPE => self::BOOLEAN),
                    ReviewPreferences::IGNORE_WHITESPACE       => array(self::TYPE => self::BOOLEAN)
                ),
                TimePreferences::TIME_PREFERENCES => array (
                    TimePreferences::DISPLAY  => array(
                        self::TYPE => self::STRING,
                        self::VALID_VALUES => array(TimePreferences::TIMEAGO, TimePreferences::TIMESTAMP)
                    )
                )
            ),
            'display_fullname' => array(self::TYPE => self::BOOLEAN),
        ),
        'jira' => array(
            'link_to_jobs'    => array(self::TYPE => self::BOOLEAN),
            'delay_job_links' => array(self::TYPE => self::INT),
            'max_job_fixes'   => array(self::TYPE => self::INT),
            'host'            => array(self::TYPE => self::HTTP_STRING),
            'user'            => array(self::TYPE => self::STRING),
            'password'        => array(self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE),
            'job_field'       => array(self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE)
        ),
        IWorkflow::WORKFLOW => array(
            IWorkflow::WORKFLOW_ENABLED => array(self::TYPE => self::BOOLEAN)
        ),
        IWorkflow::WORKFLOW_RULES => array(
            IWorkflow::ON_SUBMIT => array(
                IWorkflow::WITH_REVIEW => array(
                    IWorkflow::RULE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::NO_CHECKING, IWorkflow::APPROVED, IWorkflow::STRICT)
                    ),
                    IWorkflow::MODE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY)
                    )
                ),
                IWorkflow::WITHOUT_REVIEW => array(
                    IWorkflow::RULE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::NO_CHECKING, IWorkflow::AUTO_CREATE, IWorkflow::REJECT)
                    ),
                    IWorkflow::MODE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY)
                    )
                )
            ),
            IWorkflow::END_RULES => array(
                IWorkflow::UPDATE => array(
                    IWorkflow::RULE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::NO_CHECKING, IWorkflow::NO_REVISION)
                    ),
                    IWorkflow::MODE => array(
                        self::TYPE         => self::STRING,
                        self::VALID_VALUES => array(IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY)
                    )
                )
            ),
            IWorkflow::AUTO_APPROVE => array(
                IWorkflow::RULE => array(
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => array(IWorkflow::NEVER, IWorkflow::VOTES)
                ),
                IWorkflow::MODE => array(
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => array(IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY)
                )
            ),
            IWorkflow::COUNTED_VOTES => array(
                IWorkflow::RULE => array(
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => array(IWorkflow::ANYONE, IWorkflow::MEMBERS)
                ),
                IWorkflow::MODE => array(
                    self::TYPE         => self::STRING,
                    self::VALID_VALUES => array(IWorkflow::MODE_DEFAULT, IWorkflow::MODE_POLICY)
                )
            )
        ),
        'saml' => array('header' => array(self::TYPE => self::STRING, self::CASE_SENSITIVITY => self::CASE_SENSITIVE)),
        'mentions' => array(
            'usersBlacklist' => array(self::TYPE => self::ARRAY_OF_STRINGS, self::FORCE_STRING => true)
        ),
        self::P4 => array(
            self::SSO_ENABLED => array(self::TYPE => self::BOOLEAN),
            P4_SERVER_ID => array(self::SSO_ENABLED => array(self::TYPE => self::BOOLEAN))
        ),
        self::ENVIRONMENT => array(
            self::LOGOUT_URL =>
                array(self::TYPE => self::STRING,
                      self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                      self::ALLOW_NULL => true
                ),
            self::VENDOR => array(
                self::EMOJI_PATH => array(
                    self::TYPE => self::STRING,
                    self::CASE_SENSITIVITY => self::CASE_SENSITIVE,
                    self::ALLOW_NULL => true
                ),
            )
        ),
        self::LOG => array(
            self::REFERENCE_ID => array(self::TYPE => self::BOOLEAN),
            self::EVENT_TRACE  => array(self::TYPE => self::BOOLEAN)
        ),
        self::MARKDOWN => [
            self::FILE_EXTENSIONS => [self::TYPE => self::ARRAY_OF_STRINGS, self::FORCE_STRING => true],
            self::MARKDOWN        => [
                self::TYPE => self::STRING, self::VALID_VALUES => MarkdownSettings::MARKDOWN_OPTIONS
            ]
        ]
    );

    /**
     * Gets a value from the config and checks if it is valid. Simple case at the moment - we may
     * want to enhance to add ranges etc.
     * @param $config array config data
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @param null $default default value that will be returned without validation if any problems occur
     * @param null $metaData metadata override
     * @return array|bool|int|mixed|null|string
     * @throws ConfigException
     * @throws \Exception
     */
    public static function getValue($config, $path, $default = null, $metaData = null)
    {
        // This will throw an exception if the value is not set
        $configValue     = null;
        $metaDataElement = null;
        try {
            $configValue = self::getValueFromConfig($config, $path);
            try {
                $metaDataElement = self::getValueFromConfig($metaData ? $metaData : self::$configMetaData, $path);
            } catch (ConfigException $e) {
                // Something is wrong with our metadata, perhaps a path is asked for that we haven't yet catered for.
                // Don't fail because of this.
                return $configValue;
            }
            // This will throw an exception if validation fails
            $configValue = self::validate($configValue, $metaDataElement, $path);
        } catch (ConfigException $ce) {
            if ($default === null || $metaDataElement === null) {
                throw $ce;
            } else {
                $configValue = self::validateDefault($default, $metaDataElement);
            }
        }
        return $configValue;
    }

    /**
     * Make sure that if a default is provided it fits the metadata definition.
     * @param $value
     * @param $metaDataValue
     * @return null
     * @throws ConfigException
     */
    private static function validateDefault($value, $metaDataValue)
    {
        $type    = $metaDataValue[self::TYPE];
        $default = null;
        switch ($type) {
            case self::STRING:
                if (isset($metaDataValue[self::VALID_VALUES]) &&
                    !array_uintersect(array($value), $metaDataValue[self::VALID_VALUES], 'strcasecmp')) {
                    throw new ConfigException(
                        "Value '" . $value . "' provided as a default must be one of '"
                        . implode(', ', $metaDataValue[self::VALID_VALUES]) . '\''
                    );
                }
                $default = isset($metaDataValue[self::CASE_SENSITIVITY]) ? $value : strtolower($value);
                break;
            default:
                $default = $value;
                break;
        }
        return $default;
    }

    /**
     * Validates the value.
     * @param $value mixed value to validate
     * @param array $metaDataValue string meta data
     * @param $path string the path
     * @return array|bool|int|null|string
     * @throws ConfigException
     */
    private static function validate($value, array $metaDataValue, $path)
    {
        // Validate all parameters here - currently just type is specified
        $type           = isset($metaDataValue[self::TYPE]) ? $metaDataValue[self::TYPE] : null;
        $convertedValue = null;
        $nullPermitted  =
            $value === null && isset($metaDataValue[self::ALLOW_NULL]) && $metaDataValue[self::ALLOW_NULL];
        switch ($type) {
            case self::INT:
                if (ctype_digit(strval($value))) {
                    $convertedValue = intval($value);
                }
                break;
            case self::BOOLEAN:
                if ($value === true || strcasecmp($value, 'true') == 0) {
                    $convertedValue = true;
                } elseif ($value === false || strcasecmp($value, 'false') == 0) {
                    $convertedValue = false;
                }
                break;
            case self::STRING:
                if (is_string($value)) {
                    if (isset($metaDataValue[self::CASE_SENSITIVITY])
                        && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                        $convertedValue = $value;
                    } else {
                        $convertedValue = strtolower($value);
                    }
                    if (isset($metaDataValue[self::VALID_VALUES]) &&
                        !array_uintersect(array($value), $metaDataValue[self::VALID_VALUES], 'strcasecmp')) {
                        // valid values are specified and the value does not match them
                        $convertedValue = null;
                    }
                }
                break;
            case self::ARRAY_OF_STRINGS:
                // We support setting a single string value that we will convert to an array
                $value = is_string($value) ? (array) $value : $value;
                if (is_array($value)) {
                    $convertedValue = array();
                    foreach ($value as $arrayValue) {
                        if (isset($metaDataValue[self::FORCE_STRING])) {
                            $arrayValue = (string)$arrayValue;
                        }
                        if (is_string($arrayValue)) {
                            // Check if the case sensitivity enabled.
                            if (isset($metaDataValue[self::CASE_SENSITIVITY])
                                && $metaDataValue[self::CASE_SENSITIVITY] === self::CASE_SENSITIVE) {
                                $convertedValue[] = $arrayValue;
                                // Now also compare the array in case sensitive mode.
                                if (isset($metaDataValue[self::VALID_VALUES]) &&
                                    !array_intersect(array($arrayValue), $metaDataValue[self::VALID_VALUES])) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            } else {
                                $convertedValue[] = strtolower($arrayValue);
                                if (isset($metaDataValue[self::VALID_VALUES]) &&
                                    !array_uintersect(
                                        array($arrayValue),
                                        $metaDataValue[self::VALID_VALUES],
                                        'strcasecmp'
                                    )
                                ) {
                                    // valid values are specified and the value does not match them
                                    $convertedValue = null;
                                    break;
                                }
                            }
                        } else {
                            $convertedValue = null;
                            break;
                        }
                    }
                }
                break;
            case self::HTTP_STRING:
                if (is_string($value)) {
                    $convertedValue = rtrim($value, '/');
                    if ($convertedValue && strpos(strtolower($convertedValue), 'http') !== 0) {
                        $convertedValue = 'http://' . $convertedValue;
                    }
                }
                break;
            case self::SWARM_SETTING:
                if (array_uintersect(array($value), self::$swarmSettings, 'strcasecmp')) {
                    // SWARM_SETTING has its own set of permitted values in configuration
                    // that are case sensitive
                    $convertedValue = $value;
                } else {
                    $convertedValue = null;
                }
                break;

            case null:
                $convertedValue = $value;
                break;
        }
        if ($convertedValue === null && !$nullPermitted) {
            throw new ConfigException(
                "Value '" . (is_array($value) ? var_export($value, true) : $value) .
                "' at path '" . $path . "' is invalid"
            );
        }
        return $convertedValue;
    }

    /**
     * Iterates through the configuration to find a value.
     * @param $config array config
     * @param $path string array path (for nested values use dot notation, for example 'reviews.expand_all_file_limit'.
     * @return mixed
     * @throws ConfigException if the path does not exist or the configuration being searched is not an array.
     */
    private static function getValueFromConfig($config, $path)
    {
        $ref  = &$config;
        $keys = explode('.', $path);
        foreach ($keys as $idx => $key) {
            if (!is_array($ref)) {
                throw new ConfigException('Configuration is not an array');
            }
            if (!array_key_exists($key, $ref)) {
                throw new ConfigException("Path '" . $path . "' does not exist", ConfigException::PATH_DOES_NOT_EXIST);
            }
            $ref = &$ref[$key];
        }
        return $ref;
    }

    /**
     * Updates the merged config by overriding it with a value from the config.php file if it was set.
     * This gives the ability to replace values in merged config rather than merging them in. For example
     * Zend merges arrays so default settings of [1, 2, 3] in a module.config.php and [4] in the glob file
     * result in a merged value of [1, 2, 3, 4] when in some cases we just want [4].
     * @param ConfigMergerInterface $configListener the config listener
     * @param string                $path           string array path (for nested values use dot notation,
     *                                              for example 'reviews.expand_all_file_limit'.
     * @param string                $globConfigPath config glob path
     * @throws \ReflectionException
     * @throws \Exception if the path does not exist or the configuration being searched is not an array.
     */
    public static function updateMerged(ConfigMergerInterface $configListener, $path, $globConfigPath)
    {
        // Using reflection is not ideal but overriding the ConfigListener to access values
        // has not proved possible with the current version of Zend
        $reflection = new \ReflectionClass('\Zend\ModuleManager\Listener\ConfigListener');
        $var        = $reflection->getProperty('configs');
        $var->setAccessible(true);
        $configs = $var->getValue($configListener);
        if (isset($configs[$globConfigPath])) {
            $mergedConfig = $configListener->getMergedConfig(false);
            // If path value is set in the config.php file then use that rather than the value
            // created from merging Module/module.config.php with the glob file
            $globConfig = $configs[$globConfigPath];
            try {
                $value = self::getValueFromConfig($globConfig, $path);
                if ($value && !empty($value)) {
                    $keys = explode('.', $path);
                    $temp = &$mergedConfig;
                    foreach ($keys as $key) {
                        $temp = &$temp[$key];
                    }
                    $temp = $value;
                    $configListener->setMergedConfig($mergedConfig);
                }
            } catch (ConfigException $e) {
                // Path was not set in the glob file, will fall back to defaults
            }
        }
    }
}
